import java.lang.*;
import java.util.*;
import javax.comm.*;
import java.io.*;

public class x10Device {
	//
	// OVERVIEW
	//   x10Device is an abstraction for communicating with the FireCracker
	// proprietary serial port interface.
	//
	// USAGE
	//   Instatiate one x10Device object for each of your x10 enabled appliances.
	// Control the appliance by calling on(), off(), bright() and dim() methods
	// on the object.
	//
	// REPRESENTATION
	//   The device is represented by its x10 address (i.e. A2, J3, M12) and a
	// serial port identifier (i.e. COM1, COM2).  The device state is maintained
	// in the status variable.  It can be any multiple of 5 between 0 and 100.
	// 0 represents the off state, and 100 the full on state.  Each value in
	// between is some dim/bright combination.  An additional value of -5 is
	// maintained to represent the 'unknown' state.
	//
	String deviceAddress;
	String deviceCommand;
	CommPortIdentifier portId;
	int status;
	x10DeviceThread threadHelper;

	public x10Device(String port, String address) throws NoSuchPortException {
		// requires: address is a valid x10 address
		// modifies: this
		// effects: creates a new x10Device on serial port port with
		//   x10 address address.

		status = -5;  // device state is unknown, initially
		deviceAddress = address.toUpperCase();
		threadHelper = new x10DeviceThread(this);

		try {
			System.out.println("Attempting to open " + port.toUpperCase());
			portId = CommPortIdentifier.getPortIdentifier(port.toUpperCase());
		}
		catch (Exception e) {
			System.out.println("Error opening port " + port.toUpperCase());
		}
	}

	public void toggle() throws IOException {
		// requires: deviceAddress contains a valid address
		// modifies: this, FireCracker interface, appliance with address
		//   deviceAddress
		// effects: sets status = |100 - status|, toggles the appliance
		//   associated with deviceAddress.  if device does not have 0
		//   or 100 status, no effect.  if device is in unknown state
		//   (-5), defaults to turning device on.

		if (status <= 0) {
			setCommandValue(getCommand(deviceAddress, "ON"));
			status = 100;
		}
		else if (status == 100) {
			setCommandValue(getCommand(deviceAddress, "OFF"));
			status = 0;
		}

		threadHelper.run();
	}

	public void on() throws IOException {
		// requires: deviceAddress contains a valid address
		// modifies: this, FireCracker interface, appliance with address
		//   deviceAddress
		// effects: sets status = 100, turns on the appliance associated with
		//   deviceAddress

		setCommandValue(getCommand(deviceAddress, "ON"));
		threadHelper.run();
	}

	public void off() throws IOException {
		// requires: deviceAddress contains a valid address
		// modifies: this, FireCracker interface, appliance with address
		//   deviceAddress
		// effects: sets status = 0, turns off the appliance associated with
		//   deviceAddress

		setCommandValue(getCommand(deviceAddress, "OFF"));
		threadHelper.run();
	}

	public void dim() throws IOException {
		// requires: deviceAddress contains a valid address
		// modifies: this, FireCracker interface, appliance with address
		//   deviceAddress
		// effects: sets status = status - 5, dims the appliance associated with
		//   deviceAddress by 5%

		setCommandValue(getCommand(deviceAddress, "DIM"));
		threadHelper.run();
	}

	public void bright() throws IOException {
		// requires: deviceAddress contains a valid address
		// modifies: this, FireCracker interface, appliance with address deviceAddress
		// effects: sets status = status + 5, brightens the appliance associated with
		//   deviceAddress by 5%

		setCommandValue(getCommand(deviceAddress, "BRIGHT"));
		threadHelper.run();
	}

	public String getCommandValue() {
		// effects: returns the current command to be executed

		return deviceCommand;
	}

	public void setCommandValue(String cmd) {
		// modifies: deviceCommand
		// effects: sets the command to be executed to cmd

		deviceCommand = cmd;
	}

	public void setStatus(int stat) {
		// modifies: status
		// effects: sets the device status value to stat, if
		//   stat is a value between 0 and 100, and is divisible
		//   by 5
		if ((stat <= 100) && (stat >= -5) && (stat % 5 == 0))
			status = stat;
	}

	public int getStatus() throws UnknownStatusException {
		if (status != -5)
			return status;
		else
			throw new UnknownStatusException();
	}

	public String getDeviceAddress() {
		return deviceAddress;
	}

	public CommPortIdentifier getPortId() {
		return portId;
	}

	private static void writeStream(SerialPort pt, String val) throws IOException {
		// requires: pt is an opened SerialPort, val contains only '0' and '1'
		// modifies: the serial port
		// effects: sends the bit stream represented in val to the FireCracker

		try {
			for (int i = 0; i < val.length(); i++) {
				pt.setDTR(true);
				pt.setRTS(true);
				Thread.currentThread().sleep(1);

				switch(val.charAt(i)) {
				case '0':
					pt.setRTS(false);
					break;
				case '1':
					pt.setDTR(false);
					break;
				default:
					throw new IOException();
				}
				Thread.currentThread().sleep(1);
			}
			Thread.currentThread().sleep(250);
		}
		catch (InterruptedException e) {
			throw new IOException();
		}
	}


	private static String getCommand(String addr, String command) throws IllegalArgumentException {
		// effects: returns a bit string that, when sent to the FireCracker
		//   will execute command on the appliance at addr.  throws
		//   IllegalArgumentException if there are invalid characters or
		//   values in command or addr.

		String returnVal = new String();

		// this header is the same for all commands
		returnVal = returnVal.concat("1101010110101010");

		switch (Character.toUpperCase(addr.charAt(0))) {
		case ('A'): returnVal = returnVal.concat("0110"); break;
		case ('B'): returnVal = returnVal.concat("0111"); break;
		case ('C'): returnVal = returnVal.concat("0100"); break;
		case ('D'): returnVal = returnVal.concat("0101"); break;
		case ('E'): returnVal = returnVal.concat("1000"); break;
 		case ('F'): returnVal = returnVal.concat("1001"); break;
		case ('G'): returnVal = returnVal.concat("1010"); break;
		case ('H'): returnVal = returnVal.concat("1011"); break;
		case ('I'): returnVal = returnVal.concat("1110"); break;
		case ('J'): returnVal = returnVal.concat("1111"); break;
		case ('K'): returnVal = returnVal.concat("1100"); break;
		case ('L'): returnVal = returnVal.concat("1101"); break;
		case ('M'): returnVal = returnVal.concat("0000"); break;
		case ('N'): returnVal = returnVal.concat("0001"); break;
		case ('O'): returnVal = returnVal.concat("0010"); break;
		case ('P'): returnVal = returnVal.concat("0011"); break;
		default:
			throw new IllegalArgumentException();
		}

		if ((new Integer(addr.substring(1))).intValue() < 9)
			returnVal = returnVal.concat("0000");
		else
			returnVal = returnVal.concat("0100");

		if (command.toUpperCase().equals("ON")) {
			switch((new Integer(addr.substring(1))).intValue()) {
			case (1): returnVal = returnVal.concat("00000000"); break;
			case (2): returnVal = returnVal.concat("00010000"); break;
			case (3): returnVal = returnVal.concat("00001000"); break;
			case (4): returnVal = returnVal.concat("00011000"); break;
			case (5): returnVal = returnVal.concat("01000000"); break;
			case (6): returnVal = returnVal.concat("01010000"); break;
			case (7): returnVal = returnVal.concat("01001000"); break;
			case (8): returnVal = returnVal.concat("01011000"); break;
			case (9): returnVal = returnVal.concat("00000000"); break;
			case (10): returnVal = returnVal.concat("00010000"); break;
			case (11): returnVal = returnVal.concat("00001000"); break;
			case (12): returnVal = returnVal.concat("00011000"); break;
			case (13): returnVal = returnVal.concat("01000000"); break;
			case (14): returnVal = returnVal.concat("01010000"); break;
			case (15): returnVal = returnVal.concat("01001000"); break;
			case (16): returnVal = returnVal.concat("01011000"); break;
			}
		}
		else if (command.toUpperCase().equals("OFF")) {
			switch((new Integer(addr.substring(1))).intValue()) {
			case (1): returnVal = returnVal.concat("00100000"); break;
			case (2): returnVal = returnVal.concat("00110000"); break;
			case (3): returnVal = returnVal.concat("00101000"); break;
			case (4): returnVal = returnVal.concat("00111000"); break;
			case (5): returnVal = returnVal.concat("01100000"); break;
			case (6): returnVal = returnVal.concat("01110000"); break;
			case (7): returnVal = returnVal.concat("01101000"); break;
			case (8): returnVal = returnVal.concat("01111000"); break;
			case (9): returnVal = returnVal.concat("00100000"); break;
			case (10): returnVal = returnVal.concat("00110000"); break;
			case (11): returnVal = returnVal.concat("00011000"); break;
			case (12): returnVal = returnVal.concat("00111000"); break;
			case (13): returnVal = returnVal.concat("01100000"); break;
			case (14): returnVal = returnVal.concat("01110000"); break;
			case (15): returnVal = returnVal.concat("01101000"); break;
			case (16): returnVal = returnVal.concat("01111000"); break;
			}
		}
		else if (command.toUpperCase().equals("DIM")) {
			returnVal = returnVal.concat("10011000");
		}
		else if (command.toUpperCase().equals("BRIGHT")) {
			returnVal = returnVal.concat("10001000");
		}
		else
			throw new IllegalArgumentException();

		// the footer is the same for all commands
		return returnVal.concat("10101101");
	}
}


class x10DeviceThread extends Thread {

	x10Device parent;

	x10DeviceThread(x10Device dad) {
		super();
		parent = dad;

		// hacked by Max to try to get around
		//    "name can't be null while loading driver  com.sun.comm.Win32Driver"
		//
		System.out.println(" setting name.. ");
		setName("commthread");
		// end of Maxhack

	}

	public void run() {
		doWork();
	}

	void doWork() {

		SerialPort serialPort;

		try {
			serialPort = (SerialPort) parent.getPortId().open("x10Device.threadBaby", 2000);
			Thread.currentThread().sleep(50);  // not sure why I need this...

			writeStream(serialPort, parent.getCommandValue());
			serialPort.close();
		}
		catch (Exception e) {
			System.out.println("error in threadHelper.doWork()");
			e.printStackTrace();
			//throw new IOException();
		}
	}

	private static void writeStream(SerialPort pt, String val) throws IOException {
		// requires: pt is an opened SerialPort, val contains only '0' and '1'
		// modifies: the serial port
		// effects: sends the bit stream represented in val to the FireCracker

		try {
			for (int i = 0; i < val.length(); i++) {
				pt.setDTR(true);
				pt.setRTS(true);
				Thread.currentThread().sleep(1);

				switch(val.charAt(i)) {
				case '0':
					pt.setRTS(false);
					break;
				case '1':
					pt.setDTR(false);
					break;
				default:
					throw new IOException();
				}
				Thread.currentThread().sleep(1);
			}
			Thread.currentThread().sleep(250);
		}
		catch (InterruptedException e) {
			throw new IOException();
		}
	}
}
